import { createMachine, assign, spawn, actions, send, interpret } from 'xstate';
import { createRunner, OBJ, ARR, FUNC } from 'ploson';
import * as R from 'ramda';

import { Rule } from './rule-dm';
import { configuredCollectionDiff, idUniquesInFirst } from './utils';
import {
  getAddedRule,
  getAddedTrigger,
  createPrepareAddedRule,
  createPrepareAddedTrigger,
  // spawnTriggerActors,
  prepareUpdatedRule,
  prepareUpdatedTrigger,
  getUpdateRuleActions,
  updateTriggerAction,
} from './manager-logic';
import { runPlosonOrReject, groupByHasCallback, populateCallbacks } from './processing';

const { pure, stop } = actions;

const deferredActions = {
  ADD_RULE: { actions: 'defer' },
  ADD_TRIGGER: { actions: 'defer' },
  UPDATE_RULE: { actions: 'defer' },
  UPDATE_TRIGGER: { actions: 'defer' },
  DELETE_RULE: { actions: 'defer' },
  DELETE_TRIGGER: { actions: 'defer' },
};

const context = {
  $queue: [],
  rules: [
    // { ploson, ref, ...rule }
  ],
  triggers: [
    // { id, trigger, targets, ploson, callback, ref }
  ],
};

const machineConf = {
  predictableActionArguments: true,
  id: 'rule-manager',
  initial: 'initRules',
  context,
  states: {
    retryRules: { after: { RETRY_DELAY: { target: 'initRules' } } },
    initRules: {
      invoke: {
        src: 'ruleFetcher',
        onDone: 'prepareRules',
        onError: { actions: 'logError', target: 'retryRules' },
      },
    },
    prepareRules: {
      invoke: {
        id: 'processInitRules',
        src: 'processRules',
        onDone: { actions: 'addRules', target: 'initTriggers' },
        onError: { actions: 'logError', target: 'initTriggers' },
      },
    },
    retryTriggers: { after: { RETRY_DELAY: { target: 'initTriggers' } } },
    initTriggers: {
      invoke: {
        src: 'triggerFetcher',
        onDone: { target: 'prepareTriggers' },
        onError: { actions: 'logError', target: 'retryTriggers' },
      },
    },
    prepareTriggers: {
      invoke: {
        id: 'processInitTriggers',
        src: 'processTriggers',
        onDone: { actions: ['addTriggers'], target: 'running' },
        onError: { actions: 'logError', target: 'running' },
      },
    },
    running: {
      entry: 'spawnUpdater',
      initial: 'idle',
      on: {
        FIRE_RULES: { actions: 'fireRules' },
      },
      states: {
        idle: {
          entry: 'recallFromQueue',
          on: {
            UPDATE_RULES: { target: 'updatingRules' },
            UPDATE_TRIGGERS: { target: 'processingTriggers' },
            ADD_RULE: { target: 'prepareAddingRule' },
            ADD_TRIGGER: { target: 'prepareAddingTrigger' },
            UPDATE_RULE: { target: 'prepareUpdatingRule' },
            UPDATE_TRIGGER: { target: 'prepareUpdatingTrigger' },
            DELETE_RULE: { actions: 'removeRule' },
            DELETE_TRIGGER: { actions: 'removeTrigger' },
          },
        },
        prepareAddingRule: {
          invoke: {
            src: 'prepareAddedRule',
            onError: { actions: 'logError', target: 'idle' },
            onDone: { actions: 'addRuleExternal', target: 'idle' },
          },
          on: deferredActions,
        },
        prepareAddingTrigger: {
          invoke: {
            src: 'prepareAddedTrigger',
            onError: { actions: 'logError', target: 'idle' },
            onDone: { actions: 'addTriggerExternal', target: 'idle' },
          },
          on: deferredActions,
        },
        prepareUpdatingRule: {
          invoke: {
            src: 'prepareUpdatedRule',
            onError: { actions: 'logError', target: 'idle' },
            onDone: { actions: 'updateRuleExternal', target: 'idle' },
          },
          on: deferredActions,
        },
        prepareUpdatingTrigger: {
          invoke: {
            src: 'prepareUpdatedTrigger',
            onError: { actions: 'logError', target: 'idle' },
            onDone: { actions: 'updateTriggerExternal', target: 'idle' },
          },
          on: deferredActions,
        },
        updatingRules: {
          invoke: {
            id: 'processRuleUpdates',
            src: 'processRules',
            onDone: {
              actions: ['removeRules', 'addRules', 'updateRules'],
              target: 'idle',
            },
            onError: { actions: 'logError', target: 'idle' },
          },
        },
        processingTriggers: {
          invoke: {
            id: 'processTriggerUpdates',
            src: 'processTriggers',
            onDone: {
              actions: ['killActors', 'sendUpdateTriggers'],
              // target: 'updatingTriggers',
            },
            onError: { actions: 'logError', target: 'idle' },
          },
          on: {
            PASS_ON_RESULT: 'updatingTriggers',
          },
        },
        updatingTriggers: {
          entry: ['removeTriggers', 'updateTriggers', 'addTriggers'],
          always: { target: 'idle' },
        },
      },
    },
  },
};

const getOptions = ({
  ruleFetcher = () => Promise.resolve(ARR),
  triggerFetcher = () => Promise.resolve(ARR),
  staticEnv = OBJ,
  staticEnvTriggers = staticEnv,
  updater = FUNC,
}) => {
  const rulesDiff = configuredCollectionDiff({
    ignoreKeys: ['ref', 'ploson', 'active'],
    // skipIf: ({ active }) => !active,
    autoRemoveCondition: ({ active }) => !active,
    transformAdded: (item) => {
      const { value, error, warning } = Rule.validate(item);
      if (error || warning) console.warn(error || warning);
      const ploson = createRunner({ staticEnv });
      return { ...value, ploson };
    },
    transformUpdated: (item, { ploson, ref }) => ({ ...item, ploson, ref }),
  });

  const triggersDiff = configuredCollectionDiff({
    ignoreKeys: ['ploson', 'callback', 'ref'],
    autoRemoveCondition: ({ active }) => !active,
    transformAutoRemoved: (item, { ref }) => ({ ...item, ref }),
    transformAdded: ({ targets = ARR, ...item }) => {
      const ploson = createRunner({ staticEnv: staticEnvTriggers });
      return { ...item, targets, ploson };
    },
    transformUpdated: (item, { ploson, callback, ref }, patch) => ({
      ...item,
      ploson,
      ref,
      ...(R.any(R.propEq('path', '/trigger'))(patch) || !item.active ? OBJ : { callback }),
    }),
  });

  const prepareAddedRule = createPrepareAddedRule(staticEnv);
  const prepareAddedTrigger = createPrepareAddedTrigger(staticEnvTriggers);

  return {
    delays: { RETRY_DELAY: 30_000 },

    services: {
      ruleFetcher,
      triggerFetcher,
      processRules: async ({ rules }, { data = ARR }) => {
        const { added, updated, removed } = rulesDiff(rules)(data);
        await Promise.all(
          R.map(({ onLoad, ploson }) => {
            if (!onLoad) return Promise.resolve();
            return runPlosonOrReject({ ploson, code: onLoad });
          })(added),
        );
        // console.log('Rules processed (onLoad) and ready to be run:', added, removed, updated);
        return { added, toRemove: removed, toUpdate: updated };
      },
      processTriggers: async ({ triggers }, { data = ARR }) => {
        // console.log('processTriggers - prev, next:', triggers, data);
        const { added, updated, removed } = triggersDiff(triggers)(data);
        // console.log('processTriggers - added, updated, removed:', added, updated, removed);
        // console.log('processTriggers - added.trigger:', added?.[0]?.trigger);
        const { updatedHard = ARR, updatedSoft = ARR } = groupByHasCallback(updated);
        // console.log('processTriggers - grouped hard, soft:', updatedHard, updatedSoft);
        const toAdd = await Promise.all(populateCallbacks([...added, ...updatedHard]));
        return { added: toAdd, toRemove: removed, toUpdate: updatedSoft, toRestart: updatedHard };
      },
      prepareAddedRule,
      prepareAddedTrigger,
      prepareUpdatedRule,
      prepareUpdatedTrigger,
    },

    actions: {
      defer: assign({
        $queue: ({ $queue }, e) => R.append(e, $queue),
      }),
      recallFromQueue: pure(({ $queue: [first, ...rest] }) => [
        ...(first ? [send(first)] : []),
        assign({ $queue: rest }),
      ]),
      addRuleExternal: assign({
        rules: ({ rules }, { data }) => R.append(getAddedRule(data))(rules),
      }),
      removeRule: pure(({ rules }, { data: { id } = OBJ }) => {
        const rIx = R.findIndex(R.propEq('id', id), rules);
        if (rIx === -1) return [];
        const { ref } = rules[rIx];
        return [stop(ref), assign({ rules: R.remove(rIx, 1)(rules) })];
      }),
      addTriggerExternal: assign({
        triggers: ({ triggers }, { data }) => R.append(getAddedTrigger(data))(triggers),
      }),
      removeTrigger: pure(({ triggers }, { data: { id } = OBJ }) => {
        const rIx = R.findIndex(R.propEq('id', id), triggers);
        if (rIx === -1) return [];
        const { ref } = triggers[rIx];
        return [stop(ref), assign({ triggers: R.remove(rIx, 1)(triggers) })];
      }),
      updateRuleExternal: pure(getUpdateRuleActions),
      updateTriggerExternal: pure(updateTriggerAction),
      logError: (ctx, { data }) => console.error(data),

      killActors: pure((ctx, { data: { toRemove = ARR, toRestart = ARR } = OBJ }) => {
        // console.log('killActors toRemove, toRestart:', toRemove, toRestart);
        const events = R.pipe(
          R.filter(({ ref }) => Boolean(ref)),
          R.map(({ ref }) => stop(ref)),
        )([...toRemove, ...toRestart]);
        // console.log('killActors events:', events);
        return events;
      }),
      sendUpdateTriggers: send((ctx, { data }) => ({ type: 'PASS_ON_RESULT', data })),
      removeTriggers: assign({
        triggers: ({ triggers }, { data: { toRemove = ARR } = OBJ }) =>
          idUniquesInFirst(triggers, toRemove),
        // TODO: Destructors??
      }),
      addTriggers: assign({
        triggers: ({ triggers }, { data: { added = ARR, toRestart = ARR } = OBJ }) => {
          // console.info('Spawning triggers:', R.pluck('id', added));
          const addedTriggers = R.map(getAddedTrigger)([...added, ...toRestart]);
          return [...triggers, ...addedTriggers];
        },
      }),
      updateTriggers: assign({
        triggers: ({ triggers }, { data: { toUpdate = ARR } = OBJ }) => {
          // R.over(R.lensIndex(R.findIndex(R.propEq('id', id))), R.assoc('targets'))(triggers);
          return R.map((item) => {
            const updated = R.find(R.propEq('id', item.id))(toUpdate);
            return updated ? R.assoc('targets', updated.targets ?? [], item) : item;
          })(triggers);
        },
      }),
      removeRules: assign({
        rules: ({ rules }, { data: { toRemove = ARR } = OBJ }) => idUniquesInFirst(rules, toRemove),
      }),
      addRules: assign({
        rules: ({ rules }, { data: { added = ARR } = OBJ }) => {
          const addedRules = R.map(getAddedRule)(added);
          return [...rules, ...addedRules];
        },
      }),
      updateRules: pure((ctx, { data: { toUpdate = ARR } = OBJ }) =>
        R.map((rule) => send({ type: 'UPDATE_RULE', data: rule }))(toUpdate),
      ),
      fireRules: pure(({ rules }, { targets, payload }) => {
        // console.log('fireRules...:', targets, payload);
        const events = R.reduce(
          (acc, { id, active, ref }) =>
            active && ref && R.includes(id, targets)
              ? [...acc, send({ type: 'FIRE', payload }, { to: ref })]
              : acc,
          [],
        )(rules);
        // console.log('fireRules:', events);
        return events;
      }),
      spawnUpdater: assign({ updaterRef: () => spawn(updater, 'rule-trigger-updater') }),
    },
  };
};

export const createRuleEngineMachine = (config) => {
  return createMachine(machineConf, getOptions(config));
};

export const createRuleEngine = (...args) => interpret(createRuleEngineMachine(...args));
